@@ -42,9 +42,11 @@
 #include "qv4objectproto_p.h"
 #include "qv4mm_p.h"
 #include "qv4qobjectwrapper_p.h"
+#include "qv4identifiertable_p.h"
 #include <QtCore/qalgorithms.h>
 #include <QtCore/private/qnumeric_p.h>
 #include <QtCore/qloggingcategory.h>
+#include <private/qv4alloca_p.h>
 #include <qqmlengine.h>
 #include "PageReservation.h"
 #include "PageAllocation.h"
@@ -58,8 +60,10 @@
 #include <iostream>
 #include <cstdlib>
 #include <algorithm>
-#include "qv4alloca_p.h"
 #include "qv4profiling_p.h"
+#include "qv4mapobject_p.h"
+#include "qv4setobject_p.h"
+#include "qv4writebarrier_p.h"
 
 //#define MM_STATS
 
@@ -90,8 +94,6 @@
 #include <pthread_np.h>
 #endif
 
-#define MIN_UNMANAGED_HEAPSIZE_GC_LIMIT std::size_t(128 * 1024)
-
 Q_LOGGING_CATEGORY(lcGcStats, "qt.qml.gc.statistics")
 Q_DECLARE_LOGGING_CATEGORY(lcGcStats)
 Q_LOGGING_CATEGORY(lcGcAllocatorStats, "qt.qml.gc.allocatorStats")
@@ -110,7 +112,11 @@ enum {
 
 struct MemorySegment {
     enum {
+#ifdef Q_OS_RTEMS
+        NumChunks = sizeof(quint64),
+#else
         NumChunks = 8*sizeof(quint64),
+#endif
         SegmentSize = NumChunks*Chunk::ChunkSize,
     };
 
@@ -198,7 +204,7 @@ Chunk *MemorySegment::allocate(size_t size)
         // chunk allocated for one huge allocation
         Q_ASSERT(availableBytes >= size);
         pageReservation.commit(base, size);
-        allocatedMap = ~static_cast<quintptr>(0);
+        allocatedMap = ~static_cast<quint64>(0);
         return base;
     }
     size_t requiredChunks = (size + sizeof(Chunk) - 1)/sizeof(Chunk);
@@ -217,7 +223,8 @@ Chunk *MemorySegment::allocate(size_t size)
             pageReservation.commit(candidate, size);
             for (uint i = 0; i < requiredChunks; ++i)
                 setBit(candidate - base + i);
-            DEBUG << "allocated chunk " << candidate << hex << size;
+            DEBUG << "allocated chunk " << candidate << Qt::hex << size;
+
             return candidate;
         }
     }
@@ -335,7 +342,7 @@ bool Chunk::sweep(ExecutionEngine *engine)
 
             HeapItem *itemToFree = o + index;
             Heap::Base *b = *itemToFree;
-            const VTable *v = b->vtable();
+            const VTable *v = b->internalClass->vtable;
 //            if (Q_UNLIKELY(classCountPtr))
 //                classCountPtr(v->className);
             if (v->destroy) {
@@ -389,8 +396,8 @@ void Chunk::freeAll(ExecutionEngine *engine)
 
             HeapItem *itemToFree = o + index;
             Heap::Base *b = *itemToFree;
-            if (b->vtable()->destroy) {
-                b->vtable()->destroy(b);
+            if (b->internalClass->vtable->destroy) {
+                b->internalClass->vtable->destroy(b);
                 b->_checkIsDestroyed();
             }
 #ifdef V4_USE_HEAPTRACK
@@ -412,10 +419,6 @@ void Chunk::resetBlackBits()
     memset(blackBitmap, 0, sizeof(blackBitmap));
 }
 
-#ifdef MM_STATS
-static uint nGrayItems = 0;
-#endif
-
 void Chunk::collectGrayItems(MarkStack *markStack)
 {
     //    DEBUG << "sweeping chunk" << this << (*freeList);
@@ -438,10 +441,6 @@ void Chunk::collectGrayItems(MarkStack *markStack)
             Heap::Base *b = *itemToFree;
             Q_ASSERT(b->inUse());
             markStack->push(b);
-#ifdef MM_STATS
-            ++nGrayItems;
-//            qDebug() << "adding gray item" << b << "to mark stack";
-#endif
         }
         grayBitmap[i] = 0;
         o += Chunk::Bits;
@@ -459,7 +458,7 @@ void Chunk::sortIntoBins(HeapItem **bins, uint nBins)
 #else
     const int start = 1;
 #endif
-#ifdef MM_STATS
+#ifndef QT_NO_DEBUG
     uint freeSlots = 0;
     uint allocatedSlots = 0;
 #endif
@@ -469,7 +468,7 @@ void Chunk::sortIntoBins(HeapItem **bins, uint nBins)
         if (!i)
             usedSlots |= (static_cast<quintptr>(1) << (HeaderSize/SlotSize)) - 1;
 #endif
-#ifdef MM_STATS
+#ifndef QT_NO_DEBUG
         allocatedSlots += qPopulationCount(usedSlots);
 //        qDebug() << hex << "   i=" << i << "used=" << usedSlots;
 #endif
@@ -480,13 +479,16 @@ void Chunk::sortIntoBins(HeapItem **bins, uint nBins)
             uint freeStart = i*Bits + index;
             usedSlots &= ~((static_cast<quintptr>(1) << index) - 1);
             while (!usedSlots) {
-                ++i;
-                if (i == EntriesInBitmap) {
-                    usedSlots = (quintptr)-1;
+                if (++i < EntriesInBitmap) {
+                    usedSlots = (objectBitmap[i]|extendsBitmap[i]);
+                } else {
+                    Q_ASSERT(i == EntriesInBitmap);
+                    // Overflows to 0 when counting trailing zeroes above in next iteration.
+                    // Then, all the bits are zeroes and we break.
+                    usedSlots = std::numeric_limits<quintptr>::max();
                     break;
                 }
-                usedSlots = (objectBitmap[i]|extendsBitmap[i]);
-#ifdef MM_STATS
+#ifndef QT_NO_DEBUG
                 allocatedSlots += qPopulationCount(usedSlots);
 //                qDebug() << hex << "   i=" << i << "used=" << usedSlots;
 #endif
@@ -497,7 +499,7 @@ void Chunk::sortIntoBins(HeapItem **bins, uint nBins)
             usedSlots |= (quintptr(1) << index) - 1;
             uint freeEnd = i*Bits + index;
             uint nSlots = freeEnd - freeStart;
-#ifdef MM_STATS
+#ifndef QT_NO_DEBUG
 //            qDebug() << hex << "   got free slots from" << freeStart << "to" << freeEnd << "n=" << nSlots << "usedSlots=" << usedSlots;
             freeSlots += nSlots;
 #endif
@@ -508,7 +510,7 @@ void Chunk::sortIntoBins(HeapItem **bins, uint nBins)
             bins[bin] = freeItem;
         }
     }
-#ifdef MM_STATS
+#ifndef QT_NO_DEBUG
     Q_ASSERT(freeSlots + allocatedSlots == (EntriesInBitmap - start) * 8 * sizeof(quintptr));
 #endif
 }
@@ -624,27 +626,30 @@ void BlockAllocator::sweep()
 //    qDebug() << "BlockAlloc: sweep";
     usedSlotsAfterLastSweep = 0;
 
-    auto isFree = [this] (Chunk *c) {
-        bool isUsed = c->sweep(engine);
+    auto firstEmptyChunk = std::partition(chunks.begin(), chunks.end(), [this](Chunk *c) {
+        return c->sweep(engine);
+    });
 
-        if (isUsed) {
-            c->sortIntoBins(freeBins, NumBins);
-            usedSlotsAfterLastSweep += c->nUsedSlots();
-        } else {
-            Q_V4_PROFILE_DEALLOC(engine, Chunk::DataSize, Profiling::HeapPage);
-            chunkAllocator->free(c);
-        }
-        return !isUsed;
-    };
+    std::for_each(chunks.begin(), firstEmptyChunk, [this](Chunk *c) {
+        c->sortIntoBins(freeBins, NumBins);
+        usedSlotsAfterLastSweep += c->nUsedSlots();
+    });
 
-    auto newEnd = std::remove_if(chunks.begin(), chunks.end(), isFree);
-    chunks.erase(newEnd, chunks.end());
+    // only free the chunks at the end to avoid that the sweep() calls indirectly
+    // access freed memory
+    std::for_each(firstEmptyChunk, chunks.end(), [this](Chunk *c) {
+        Q_V4_PROFILE_DEALLOC(engine, Chunk::DataSize, Profiling::HeapPage);
+        chunkAllocator->free(c);
+    });
+
+    chunks.erase(firstEmptyChunk, chunks.end());
 }
 
 void BlockAllocator::freeAll()
 {
-    for (auto c : chunks) {
+    for (auto c : chunks)
         c->freeAll(engine);
+    for (auto c : chunks) {
         Q_V4_PROFILE_DEALLOC(engine, Chunk::DataSize, Profiling::HeapPage);
         chunkAllocator->free(c);
     }
@@ -668,11 +673,10 @@ HeapItem *HugeItemAllocator::allocate(size_t size) {
     Chunk *c = nullptr;
     if (size >= MemorySegment::SegmentSize/2) {
         // too large to handle through the ChunkAllocator, let's get our own memory segement
-        size_t segmentSize = size + Chunk::HeaderSize; // space required for the Chunk header
+        size += Chunk::HeaderSize; // space required for the Chunk header
         size_t pageSize = WTF::pageSize();
-        segmentSize = (segmentSize + pageSize - 1) & ~(pageSize - 1); // align to page sizes
-        m = new MemorySegment(segmentSize);
         size = (size + pageSize - 1) & ~(pageSize - 1); // align to page sizes
+        m = new MemorySegment(size);
         c = m->allocate(size);
     } else {
         c = chunkAllocator->allocate(size);
@@ -691,7 +695,7 @@ static void freeHugeChunk(ChunkAllocator *chunkAllocator, const HugeItemAllocato
 {
     HeapItem *itemToFree = c.chunk->first();
     Heap::Base *b = *itemToFree;
-    const VTable *v = b->vtable();
+    const VTable *v = b->internalClass->vtable;
     if (Q_UNLIKELY(classCountPtr))
         classCountPtr(v->className);
 
@@ -758,10 +762,11 @@ MemoryManager::MemoryManager(ExecutionEngine *engine)
     : engine(engine)
     , chunkAllocator(new ChunkAllocator)
     , blockAllocator(chunkAllocator, engine)
+    , icAllocator(chunkAllocator, engine)
     , hugeItemAllocator(chunkAllocator, engine)
     , m_persistentValues(new PersistentValueStorage(engine))
     , m_weakValues(new PersistentValueStorage(engine))
-    , unmanagedHeapSizeGCLimit(MIN_UNMANAGED_HEAPSIZE_GC_LIMIT)
+    , unmanagedHeapSizeGCLimit(MinUnmanagedHeapSizeGCLimit)
     , aggressiveGC(!qEnvironmentVariableIsEmpty("QV4_MM_AGGRESSIVE_GC"))
     , gcStats(lcGcStats().isDebugEnabled())
     , gcCollectorStats(lcGcAllocatorStats().isDebugEnabled())
@@ -774,11 +779,6 @@ MemoryManager::MemoryManager(ExecutionEngine *engine)
         blockAllocator.allocationStats = statistics.allocations;
 }
 
-#ifdef MM_STATS
-static int allocationCount = 0;
-static size_t lastAllocRequestedSlots = 0;
-#endif
-
 Heap::Base *MemoryManager::allocString(std::size_t unmanagedSize)
 {
     const size_t stringSize = align(sizeof(Heap::String));
@@ -786,35 +786,9 @@ Heap::Base *MemoryManager::allocString(std::size_t unmanagedSize)
     lastAllocRequestedSlots = stringSize >> Chunk::SlotSizeShift;
     ++allocationCount;
 #endif
-
-    bool didGCRun = false;
-    if (aggressiveGC) {
-        runGC();
-        didGCRun = true;
-    }
-
     unmanagedHeapSize += unmanagedSize;
-    if (unmanagedHeapSize > unmanagedHeapSizeGCLimit) {
-        if (!didGCRun)
-            runGC();
 
-        if (3*unmanagedHeapSizeGCLimit <= 4*unmanagedHeapSize)
-            // more than 75% full, raise limit
-            unmanagedHeapSizeGCLimit = std::max(unmanagedHeapSizeGCLimit, unmanagedHeapSize) * 2;
-        else if (unmanagedHeapSize * 4 <= unmanagedHeapSizeGCLimit)
-            // less than 25% full, lower limit
-            unmanagedHeapSizeGCLimit = qMax(MIN_UNMANAGED_HEAPSIZE_GC_LIMIT, unmanagedHeapSizeGCLimit/2);
-        didGCRun = true;
-    }
-
-    HeapItem *m = blockAllocator.allocate(stringSize);
-    if (!m) {
-        if (!didGCRun && shouldRunGC())
-            runGC();
-        m = blockAllocator.allocate(stringSize, true);
-    }
-
-//    qDebug() << "allocated string" << m;
+    HeapItem *m = allocate(&blockAllocator, stringSize);
     memset(m, 0, stringSize);
     return *m;
 }
@@ -826,32 +800,11 @@ Heap::Base *MemoryManager::allocData(std::size_t size)
     ++allocationCount;
 #endif
 
-    bool didRunGC = false;
-    if (aggressiveGC) {
-        runGC();
-        didRunGC = true;
-    }
-
     Q_ASSERT(size >= Chunk::SlotSize);
     Q_ASSERT(size % Chunk::SlotSize == 0);
 
-//    qDebug() << "unmanagedHeapSize:" << unmanagedHeapSize << "limit:" << unmanagedHeapSizeGCLimit << "unmanagedSize:" << unmanagedSize;
-
-    if (size > Chunk::DataSize) {
-        HeapItem *h = hugeItemAllocator.allocate(size);
-//        qDebug() << "allocating huge item" << h;
-        return *h;
-    }
-
-    HeapItem *m = blockAllocator.allocate(size);
-    if (!m) {
-        if (!didRunGC && shouldRunGC())
-            runGC();
-        m = blockAllocator.allocate(size, true);
-    }
-
+    HeapItem *m = allocate(&blockAllocator, size);
     memset(m, 0, size);
-//    qDebug() << "allocating data" << m;
     return *m;
 }
 
@@ -884,7 +837,7 @@ Heap::Object *MemoryManager::allocObjectWithMemberData(const QV4::VTable *vtable
             Chunk::clearBit(c->extendsBitmap, index);
         }
         o->memberData.set(engine, m);
-        m->internalClass = engine->internalClasses[EngineBase::Class_MemberData];
+        m->internalClass.set(engine, engine->internalClasses(EngineBase::Class_MemberData));
         Q_ASSERT(o->memberData->internalClass);
         m->values.alloc = static_cast<uint>((memberSize - sizeof(Heap::MemberData) + sizeof(Value))/sizeof(Value));
         m->values.size = o->memberData->values.alloc;
@@ -898,20 +851,22 @@ Heap::Object *MemoryManager::allocObjectWithMemberData(const QV4::VTable *vtable
 static uint markStackSize = 0;
 
 MarkStack::MarkStack(ExecutionEngine *engine)
-    : engine(engine)
+    : m_engine(engine)
 {
-    base = (Heap::Base **)engine->gcStack->base();
-    top = base;
-    limit = base + ExecutionEngine::GCStackLimit/sizeof(Heap::Base)*3/4;
+    m_base = (Heap::Base **)engine->gcStack->base();
+    m_top = m_base;
+    const size_t size = engine->maxGCStackSize() / sizeof(Heap::Base);
+    m_hardLimit = m_base + size;
+    m_softLimit = m_base + size * 3 / 4;
 }
 
 void MarkStack::drain()
 {
-    while (top > base) {
+    while (m_top > m_base) {
         Heap::Base *h = pop();
         ++markStackSize;
         Q_ASSERT(h); // at this point we should only have Heap::Base objects in this area on the stack. If not, weird things might happen.
-        h->markChildren(this);
+        h->internalClass->vtable->markObjects(h, this);
     }
 }
 
@@ -954,20 +909,15 @@ void MemoryManager::collectRoots(MarkStack *markStack)
 
         if (keepAlive)
             qobjectWrapper->mark(markStack);
-
-        if (markStack->top >= markStack->limit)
-            markStack->drain();
     }
 }
 
 void MemoryManager::mark()
 {
     markStackSize = 0;
-
     MarkStack markStack(engine);
     collectRoots(&markStack);
-
-    markStack.drain();
+    // dtor of MarkStack drains
 }
 
 void MemoryManager::sweep(bool lastSweep, ClassDestroyStatsCallback classCountPtr)
@@ -980,8 +930,29 @@ void MemoryManager::sweep(bool lastSweep, ClassDestroyStatsCallback classCountPt
         // signal before we start sweeping the heap
         if (QObjectWrapper *qobjectWrapper = (*it).as<QObjectWrapper>())
             qobjectWrapper->destroyObject(lastSweep);
+    }
 
-        (*it) = Primitive::undefinedValue();
+    // remove objects from weak maps and sets
+    Heap::MapObject *map = weakMaps;
+    Heap::MapObject **lastMap = &weakMaps;
+    while (map) {
+        if (map->isMarked()) {
+            map->removeUnmarkedKeys();
+            *lastMap = map;
+            lastMap = &map->nextWeakMap;
+        }
+        map = map->nextWeakMap;
+    }
+
+    Heap::SetObject *set = weakSets;
+    Heap::SetObject **lastSet = &weakSets;
+    while (set) {
+        if (set->isMarked()) {
+            set->removeUnmarkedKeys();
+            *lastSet = set;
+            lastSet = &set->nextWeakSet;
+        }
+        set = set->nextWeakSet;
     }
 
     // onDestruction handlers may have accessed other QObject wrappers and reset their value, so ensure
@@ -990,7 +961,7 @@ void MemoryManager::sweep(bool lastSweep, ClassDestroyStatsCallback classCountPt
         Managed *m = (*it).managed();
         if (!m || m->markBit())
             continue;
-        (*it) = Primitive::undefinedValue();
+        (*it) = Value::undefinedValue();
     }
 
     // Now it is time to free QV4::QObjectWrapper Value, we must check the Value's tag to make sure its object has been destroyed
@@ -1017,24 +988,29 @@ void MemoryManager::sweep(bool lastSweep, ClassDestroyStatsCallback classCountPt
         }
     }
 
-    blockAllocator.sweep();
-    hugeItemAllocator.sweep(classCountPtr);
+
+    if (!lastSweep) {
+        engine->identifierTable->sweep();
+        blockAllocator.sweep(/*classCountPtr*/);
+        hugeItemAllocator.sweep(classCountPtr);
+        icAllocator.sweep(/*classCountPtr*/);
+    }
 }
 
 bool MemoryManager::shouldRunGC() const
 {
-    size_t total = blockAllocator.totalSlots();
+    size_t total = blockAllocator.totalSlots() + icAllocator.totalSlots();
     if (total > MinSlotsGCLimit && usedSlotsAfterLastFullSweep * GCOverallocation < total * 100)
         return true;
     return false;
 }
 
-size_t dumpBins(BlockAllocator *b, bool printOutput = true)
+static size_t dumpBins(BlockAllocator *b, const char *title)
 {
     const QLoggingCategory &stats = lcGcAllocatorStats();
     size_t totalSlotMem = 0;
-    if (printOutput)
-        qDebug(stats) << "Slot map:";
+    if (title)
+        qDebug(stats) << "Slot map for" << title << "allocator:";
     for (uint i = 0; i < BlockAllocator::NumBins; ++i) {
         uint nEntries = 0;
         HeapItem *h = b->freeBins[i];
@@ -1043,17 +1019,17 @@ size_t dumpBins(BlockAllocator *b, bool printOutput = true)
             totalSlotMem += h->freeData.availableSlots;
             h = h->freeData.next;
         }
-        if (printOutput)
+        if (title)
             qDebug(stats) << "    number of entries in slot" << i << ":" << nEntries;
     }
     SDUMP() << "    large slot map";
     HeapItem *h = b->freeBins[BlockAllocator::NumBins - 1];
     while (h) {
-        SDUMP() << "        " << hex << (quintptr(h)/32) << h->freeData.availableSlots;
+        SDUMP() << "        " << Qt::hex << (quintptr(h)/32) << h->freeData.availableSlots;
         h = h->freeData.next;
     }
 
-    if (printOutput)
+    if (title)
         qDebug(stats) << "  total mem in bins" << totalSlotMem*Chunk::SlotSize;
     return totalSlotMem*Chunk::SlotSize;
 }
@@ -1094,11 +1070,8 @@ void MemoryManager::runGC()
         size_t oldChunks = blockAllocator.chunks.size();
         qDebug(stats) << "Allocated" << totalMem << "bytes in" << oldChunks << "chunks";
         qDebug(stats) << "Fragmented memory before GC" << (totalMem - usedBefore);
-        dumpBins(&blockAllocator);
-
-#ifdef MM_STATS
-        nGrayItems = 0;
-#endif
+        dumpBins(&blockAllocator, "Block");
+        dumpBins(&icAllocator, "InternalClass");
 
         QElapsedTimer t;
         t.start();
@@ -1116,7 +1089,8 @@ void MemoryManager::runGC()
             qDebug(stats) << "   new unmanaged heap:" << unmanagedHeapSize;
             qDebug(stats) << "   unmanaged heap limit:" << unmanagedHeapSizeGCLimit;
         }
-        size_t memInBins = dumpBins(&blockAllocator);
+        size_t memInBins = dumpBins(&blockAllocator, "Block")
+                + dumpBins(&icAllocator, "InternalClasss");
         qDebug(stats) << "Marked object in" << markTime << "us.";
         qDebug(stats) << "   " << markStackSize << "objects marked";
         qDebug(stats) << "Sweeped object in" << sweepTime << "us.";
@@ -1138,7 +1112,8 @@ void MemoryManager::runGC()
         qDebug(stats) << "Used memory after GC:" << usedAfter;
         qDebug(stats) << "Freed up bytes      :" << (usedBefore - usedAfter);
         qDebug(stats) << "Freed up chunks     :" << (oldChunks - blockAllocator.chunks.size());
-        size_t lost = blockAllocator.allocatedMem() - memInBins - usedAfter;
+        size_t lost = blockAllocator.allocatedMem() + icAllocator.allocatedMem()
+                - memInBins - usedAfter;
         if (lost)
             qDebug(stats) << "!!!!!!!!!!!!!!!!!!!!! LOST MEM:" << lost << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!";
         if (largeItemsBefore || largeItemsAfter) {
@@ -1159,24 +1134,28 @@ void MemoryManager::runGC()
 
     if (aggressiveGC) {
         // ensure we don't 'loose' any memory
-        Q_ASSERT(blockAllocator.allocatedMem() == getUsedMem() + dumpBins(&blockAllocator, false));
+        Q_ASSERT(blockAllocator.allocatedMem()
+                 == blockAllocator.usedMem() + dumpBins(&blockAllocator, nullptr));
+        Q_ASSERT(icAllocator.allocatedMem()
+                 == icAllocator.usedMem() + dumpBins(&icAllocator, nullptr));
     }
 
-    usedSlotsAfterLastFullSweep = blockAllocator.usedSlotsAfterLastSweep;
+    usedSlotsAfterLastFullSweep = blockAllocator.usedSlotsAfterLastSweep + icAllocator.usedSlotsAfterLastSweep;
 
     // reset all black bits
     blockAllocator.resetBlackBits();
     hugeItemAllocator.resetBlackBits();
+    icAllocator.resetBlackBits();
 }
 
 size_t MemoryManager::getUsedMem() const
 {
-    return blockAllocator.usedMem();
+    return blockAllocator.usedMem() + icAllocator.usedMem();
 }
 
 size_t MemoryManager::getAllocatedMem() const
 {
-    return blockAllocator.allocatedMem() + hugeItemAllocator.usedMem();
+    return blockAllocator.allocatedMem() + icAllocator.allocatedMem() + hugeItemAllocator.usedMem();
 }
 
 size_t MemoryManager::getLargeItemsMem() const
@@ -1184,6 +1163,18 @@ size_t MemoryManager::getLargeItemsMem() const
     return hugeItemAllocator.usedMem();
 }
 
+void MemoryManager::registerWeakMap(Heap::MapObject *map)
+{
+    map->nextWeakMap = weakMaps;
+    weakMaps = map;
+}
+
+void MemoryManager::registerWeakSet(Heap::SetObject *set)
+{
+    set->nextWeakSet = weakSets;
+    weakSets = set;
+}
+
 MemoryManager::~MemoryManager()
 {
     delete m_persistentValues;
@@ -1193,6 +1184,7 @@ MemoryManager::~MemoryManager()
     sweep(/*lastSweep*/true);
     blockAllocator.freeAll();
     hugeItemAllocator.freeAll();
+    icAllocator.freeAll();
 
     delete m_weakValues;
 #ifdef V4_USE_VALGRIND
