@@ -60,6 +60,7 @@ QT_REQUIRE_CONFIG(quick_viewtransitions);
 #include <QtQml/qqml.h>
 #include <private/qqmlguard_p.h>
 #include <private/qquicktransition_p.h>
+#include <private/qanimationjobutil_p.h>
 
 QT_BEGIN_NAMESPACE
 
@@ -157,6 +158,7 @@ public:
     bool prepareTransition(QQuickItemViewTransitioner *transitioner, int index, const QRectF &viewBounds);
     void startTransition(QQuickItemViewTransitioner *transitioner, int index);
 
+    SelfDeletable m_selfDeletable;
     QPointF nextTransitionTo;
     QPointF lastMovedTo;
     QPointF nextTransitionFrom;
@@ -192,6 +194,10 @@ class QQuickViewTransitionAttached : public QObject
     Q_PROPERTY(QList<int> targetIndexes READ targetIndexes NOTIFY targetIndexesChanged)
     Q_PROPERTY(QQmlListProperty<QObject> targetItems READ targetItems NOTIFY targetItemsChanged)
 
+    QML_NAMED_ELEMENT(ViewTransition)
+    QML_UNCREATABLE("ViewTransition is only available via attached properties.")
+    QML_ATTACHED(QQuickViewTransitionAttached)
+
 public:
     QQuickViewTransitionAttached(QObject *parent);
 
@@ -225,6 +231,5 @@ private:
 QT_END_NAMESPACE
 
 QML_DECLARE_TYPE(QQuickViewTransitionAttached)
-QML_DECLARE_TYPEINFO(QQuickViewTransitionAttached, QML_HAS_ATTACHED_PROPERTIES)
 
 #endif // QQUICKITEMVIEWTRANSITION_P_P_H
