@@ -3,17 +3,20 @@
 
 /*---
 esid: sec-array.prototype.map
-es5id: 15.4.4.19-1-3
-description: Array.prototype.map - applied to boolean primitive
+es5id: 15.4.4.19-5-19
+description: Array.prototype.map - the Arguments object can be used as thisArg
 ---*/
 
+var arg;
+
 function callbackfn(val, idx, obj) {
-  return obj instanceof Boolean;
+  return this === arg;
 }
 
-Boolean.prototype[0] = true;
-Boolean.prototype.length = 1;
+arg = (function() {
+  return arguments;
+}(1, 2, 3));
 
-var testResult = Array.prototype.map.call(false, callbackfn);
+var testResult = [11].map(callbackfn, arg);
 
 assert.sameValue(testResult[0], true, 'testResult[0]');
