@@ -2,18 +2,19 @@
 // This code is governed by the BSD license found in the LICENSE file.
 
 /*---
-es5id: 8.12.1-1_33
+es5id: 8.12.1-1_34
 description: >
-    Properties - [[HasOwnProperty]] (configurable, enumerable own
-    setter property)
+    Properties - [[HasOwnProperty]] (non-configurable, non-enumerable
+    own getter/setter property)
 ---*/
 
 var o = {};
 Object.defineProperty(o, "foo", {
-  set: function() {;
+  get: function() {
+    return 42;
   },
-  enumerable: true,
-  configurable: true
+  set: function() {;
+  }
 });
 
 assert(o.hasOwnProperty("foo"), 'o.hasOwnProperty("foo") !== true');
