@@ -5,10 +5,10 @@
 info: |
     No matter how control leaves the embedded 'Statement',
     the scope chain is always restored to its former state
-es5id: 12.10_A3.7_T5
+es5id: 12.10_A3.5_T5
 description: >
-    Declaring and calling a function within "with" statement, leading
-    to completion by exception
+    Using "with" statement within "for-in" statement, leading to
+    completion by break
 flags: [noStrict]
 ---*/
 
@@ -22,37 +22,25 @@ var myObj = {
     valueOf : function(){return 'obj_valueOf';}
 }
 
-try {
+for(var prop in myObj){
     with(myObj){
-        (function f(){
-            throw value;
-            p1 = 'x1';
-        })();
+        break;
+        p1 = 'x1';
     }
-} catch(e){
-    result = p1;
 }
 
 //////////////////////////////////////////////////////////////////////////////
 //CHECK#1
-if(result !== 1){
-  $ERROR('#1: result === 1. Actual:  result ==='+ result  );
-}
-//
-//////////////////////////////////////////////////////////////////////////////
-
-//////////////////////////////////////////////////////////////////////////////
-//CHECK#2
 if(p1 !== 1){
-  $ERROR('#2: p1 === 1. Actual:  p1 ==='+ p1  );
+  $ERROR('#1: p1 === 1. Actual:  p1 ==='+ p1  );
 }
 //
 //////////////////////////////////////////////////////////////////////////////
 
 //////////////////////////////////////////////////////////////////////////////
-//CHECK#3
+//CHECK#1
 if(myObj.p1 !== "a"){
-  $ERROR('#3: myObj.p1 === "a". Actual:  myObj.p1 ==='+ myObj.p1  );
+  $ERROR('#1: myObj.p1 === "a". Actual:  myObj.p1 ==='+ myObj.p1  );
 }
 //
 //////////////////////////////////////////////////////////////////////////////
