@@ -32,11 +32,16 @@
 #include <QQmlEngine>
 #include <QProcess>
 #include <QLibraryInfo>
+#include <QStandardPaths>
 #include <QSysInfo>
 #include <QLoggingCategory>
 #include <private/qqmlcomponent_p.h>
+#include <private/qqmlscriptdata_p.h>
+#include <qtranslator.h>
 
-class tst_qmlcachegen: public QObject
+#include "../../shared/util.h"
+
+class tst_qmlcachegen: public QQmlDataTest
 {
     Q_OBJECT
 
@@ -50,15 +55,29 @@ private slots:
     void aheadOfTimeCompilation();
     void functionExpressions();
     void versionChecksForAheadOfTimeUnits();
+    void retainedResources();
+    void skippedResources();
 
     void workerScripts();
 
     void trickyPaths_data();
     void trickyPaths();
 
-    void scriptImport();
+    void qrcScriptImport();
+    void fsScriptImport();
+    void moduleScriptImport();
+    void esModulesViaQJSEngine();
 
     void enums();
+
+    void sourceFileIndices();
+
+    void reproducibleCache_data();
+    void reproducibleCache();
+
+    void parameterAdjustment();
+    void inlineComponent();
+    void posthocRequired();
 };
 
 // A wrapper around QQmlComponent to ensure the temporary reference counts
@@ -105,6 +124,14 @@ static bool generateCache(const QString &qmlFileName, QByteArray *capturedStderr
 void tst_qmlcachegen::initTestCase()
 {
     qputenv("QML_FORCE_DISK_CACHE", "1");
+    QStandardPaths::setTestModeEnabled(true);
+
+    // make sure there's no pre-existing cache dir
+    QString cacheDir = QStandardPaths::writableLocation(QStandardPaths::CacheLocation);
+    if (!cacheDir.isEmpty())
+        //QDir(cacheDir).removeRecursively();
+        qDebug() << cacheDir;
+    QQmlDataTest::initTestCase();
 }
 
 void tst_qmlcachegen::loadGeneratedFile()
@@ -137,6 +164,7 @@ void tst_qmlcachegen::loadGeneratedFile()
         QVERIFY(cacheUnit);
         QVERIFY(cacheUnit->flags & QV4::CompiledData::Unit::StaticData);
         QVERIFY(cacheUnit->flags & QV4::CompiledData::Unit::PendingTypeCompilation);
+        QCOMPARE(uint(cacheUnit->sourceFileIndex), uint(0));
     }
 
     QVERIFY(QFile::remove(testFilePath));
@@ -151,15 +179,31 @@ void tst_qmlcachegen::loadGeneratedFile()
     QVERIFY(componentPrivate);
     auto compilationUnit = componentPrivate->compilationUnit;
     QVERIFY(compilationUnit);
-    QVERIFY(compilationUnit->data);
-    QVERIFY(!(compilationUnit->data->flags & QV4::CompiledData::Unit::StaticData));
+    auto unitData = compilationUnit->unitData();
+    QVERIFY(unitData);
+    QVERIFY(unitData->flags & QV4::CompiledData::Unit::StaticData);
 }
 
+class QTestTranslator : public QTranslator
+{
+public:
+    QString translate(const char *context, const char *sourceText, const char */*disambiguation*/, int /*n*/) const override
+    {
+        m_lastContext = QString::fromUtf8(context);
+        return QString::fromUtf8(sourceText).toUpper();
+    }
+    bool isEmpty() const override { return true; }
+    mutable QString m_lastContext;
+};
+
 void tst_qmlcachegen::translationExpressionSupport()
 {
     QTemporaryDir tempDir;
     QVERIFY(tempDir.isValid());
 
+    QTestTranslator translator;
+    qApp->installTranslator(&translator);
+
     const auto writeTempFile = [&tempDir](const QString &fileName, const char *contents) {
         QFile f(tempDir.path() + '/' + fileName);
         const bool ok = f.open(QIODevice::WriteOnly | QIODevice::Truncate);
@@ -193,7 +237,8 @@ void tst_qmlcachegen::translationExpressionSupport()
     CleanlyLoadingComponent component(&engine, QUrl::fromLocalFile(testFilePath));
     QScopedPointer<QObject> obj(component.create());
     QVERIFY(!obj.isNull());
-    QCOMPARE(obj->property("text").toString(), QString("All Ok"));
+    QCOMPARE(obj->property("text").toString(), QString("ALL Ok"));
+    QCOMPARE(translator.m_lastContext, QStringLiteral("test"));
 }
 
 void tst_qmlcachegen::signalHandlerParameters()
@@ -223,12 +268,40 @@ void tst_qmlcachegen::signalHandlerParameters()
     QVERIFY(QFile::exists(cacheFilePath));
     QVERIFY(QFile::remove(testFilePath));
 
+    {
+        QFile cache(cacheFilePath);
+        QVERIFY(cache.open(QIODevice::ReadOnly));
+        const QV4::CompiledData::Unit *cacheUnit = reinterpret_cast<const QV4::CompiledData::Unit *>(cache.map(/*offset*/0, sizeof(QV4::CompiledData::Unit)));
+        QVERIFY(cacheUnit);
+    }
+
     QQmlEngine engine;
     CleanlyLoadingComponent component(&engine, QUrl::fromLocalFile(testFilePath));
     QScopedPointer<QObject> obj(component.create());
     QVERIFY(!obj.isNull());
     QMetaObject::invokeMethod(obj.data(), "runTest");
     QCOMPARE(obj->property("result").toInt(), 42);
+
+    {
+        auto componentPrivate = QQmlComponentPrivate::get(&component);
+        QVERIFY(componentPrivate);
+        auto compilationUnit = componentPrivate->compilationUnit;
+        QVERIFY(compilationUnit);
+        QVERIFY(compilationUnit->unitData());
+
+        // Verify that the QML objects don't come from the original data.
+        QVERIFY(compilationUnit->objectAt(0) != compilationUnit->unitData()->qmlUnit()->objectAt(0));
+
+        // Typically the final file name is one of those strings that is not in the original
+        // pre-compiled qml file's string table, while for example the signal parameter
+        // name ("value") is.
+        const auto isStringIndexInStringTable = [compilationUnit](uint index) {
+            return index < compilationUnit->unitData()->stringTableSize;
+        };
+
+        QVERIFY(isStringIndexInStringTable(compilationUnit->objectAt(0)->signalAt(0)->parameterAt(0)->nameIndex));
+        QVERIFY(!compilationUnit->dynamicStrings.isEmpty());
+    }
 }
 
 void tst_qmlcachegen::errorOnArgumentsInSignalHandler()
@@ -293,12 +366,13 @@ static QQmlPrivate::CachedQmlUnit *temporaryModifiedCachedUnit = nullptr;
 
 void tst_qmlcachegen::versionChecksForAheadOfTimeUnits()
 {
-    QVERIFY(QFile::exists(":/versionchecks.qml"));
-    QCOMPARE(QFileInfo(":/versionchecks.qml").size(), 0);
+    QVERIFY(QFile::exists(":/data/versionchecks.qml"));
+    QVERIFY(QFileInfo(":/data/versionchecks.qml").size() > 0);
 
     Q_ASSERT(!temporaryModifiedCachedUnit);
     QQmlMetaType::CachedUnitLookupError error = QQmlMetaType::CachedUnitLookupError::NoError;
-    const QV4::CompiledData::Unit *originalUnit = QQmlMetaType::findCachedCompilationUnit(QUrl("qrc:/versionchecks.qml"), &error);
+    const QV4::CompiledData::Unit *originalUnit = QQmlMetaType::findCachedCompilationUnit(
+            QUrl("qrc:/data/versionchecks.qml"), &error);
     QVERIFY(originalUnit);
     QV4::CompiledData::Unit *tweakedUnit = (QV4::CompiledData::Unit *)malloc(originalUnit->unitSize);
     memcpy(reinterpret_cast<void *>(tweakedUnit), reinterpret_cast<const void *>(originalUnit), originalUnit->unitSize);
@@ -306,7 +380,7 @@ void tst_qmlcachegen::versionChecksForAheadOfTimeUnits()
     temporaryModifiedCachedUnit = new QQmlPrivate::CachedQmlUnit{tweakedUnit, nullptr, nullptr};
 
     auto testHandler = [](const QUrl &url) -> const QQmlPrivate::CachedQmlUnit * {
-        if (url == QUrl("qrc:/versionchecks.qml"))
+        if (url == QUrl("qrc:/data/versionchecks.qml"))
             return temporaryModifiedCachedUnit;
         return nullptr;
     };
@@ -314,15 +388,14 @@ void tst_qmlcachegen::versionChecksForAheadOfTimeUnits()
 
     {
         QQmlMetaType::CachedUnitLookupError error = QQmlMetaType::CachedUnitLookupError::NoError;
-        QVERIFY(!QQmlMetaType::findCachedCompilationUnit(QUrl("qrc:/versionchecks.qml"), &error));
+        QVERIFY(!QQmlMetaType::findCachedCompilationUnit(QUrl("qrc:/data/versionchecks.qml"), &error));
         QCOMPARE(error, QQmlMetaType::CachedUnitLookupError::VersionMismatch);
     }
 
     {
         QQmlEngine engine;
-        QQmlComponent component(&engine, QUrl("qrc:/versionchecks.qml"));
-        QCOMPARE(component.status(), QQmlComponent::Error);
-        QCOMPARE(component.errorString(), QString("qrc:/versionchecks.qml:-1 File was compiled ahead of time with an incompatible version of Qt and the original file cannot be found. Please recompile\n"));
+        CleanlyLoadingComponent component(&engine, QUrl("qrc:/data/versionchecks.qml"));
+        QCOMPARE(component.status(), QQmlComponent::Ready);
     }
 
     Q_ASSERT(temporaryModifiedCachedUnit);
@@ -333,14 +406,33 @@ void tst_qmlcachegen::versionChecksForAheadOfTimeUnits()
     QQmlMetaType::removeCachedUnitLookupFunction(testHandler);
 }
 
+void tst_qmlcachegen::retainedResources()
+{
+    QFile file(":/Retain.qml");
+    QVERIFY(file.open(QIODevice::ReadOnly));
+    QVERIFY(file.readAll().startsWith("import QtQml 2.0"));
+}
+
+void tst_qmlcachegen::skippedResources()
+{
+    QFile file(":/not/Skip.qml");
+    QVERIFY(file.open(QIODevice::ReadOnly));
+    QVERIFY(file.readAll().startsWith("import QtQml 2.0"));
+
+    QQmlMetaType::CachedUnitLookupError error = QQmlMetaType::CachedUnitLookupError::NoError;
+    const auto *unit = QQmlMetaType::findCachedCompilationUnit(QUrl("qrc:/not/Skip.qml"), &error);
+    QCOMPARE(unit, nullptr);
+    QCOMPARE(error, QQmlMetaType::CachedUnitLookupError::NoUnitFound);
+}
+
 void tst_qmlcachegen::workerScripts()
 {
-    QVERIFY(QFile::exists(":/workerscripts/worker.js"));
-    QVERIFY(QFile::exists(":/workerscripts/worker.qml"));
-    QCOMPARE(QFileInfo(":/workerscripts/worker.js").size(), 0);
+    QVERIFY(QFile::exists(":/workerscripts/data/worker.js"));
+    QVERIFY(QFile::exists(":/workerscripts/data/worker.qml"));
+    QVERIFY(QFileInfo(":/workerscripts/data/worker.js").size() > 0);
 
     QQmlEngine engine;
-    CleanlyLoadingComponent component(&engine, QUrl("qrc:///workerscripts/worker.qml"));
+    CleanlyLoadingComponent component(&engine, QUrl("qrc:///workerscripts/data/worker.qml"));
     QScopedPointer<QObject> obj(component.create());
     QVERIFY(!obj.isNull());
     QTRY_VERIFY(obj->property("success").toBool());
@@ -426,7 +518,7 @@ void tst_qmlcachegen::trickyPaths()
 {
     QFETCH(QString, filePath);
     QVERIFY2(QFile::exists(filePath), qPrintable(filePath));
-    QCOMPARE(QFileInfo(filePath).size(), 0);
+    QVERIFY(QFileInfo(filePath).size() > 0);
     QQmlEngine engine;
     QQmlComponent component(&engine, QUrl("qrc" + filePath));
     QScopedPointer<QObject> obj(component.create());
@@ -434,24 +526,193 @@ void tst_qmlcachegen::trickyPaths()
     QCOMPARE(obj->property("success").toInt(), 42);
 }
 
-void tst_qmlcachegen::scriptImport()
+void tst_qmlcachegen::qrcScriptImport()
 {
     QQmlEngine engine;
-    CleanlyLoadingComponent component(&engine, QUrl("qrc:///jsimport.qml"));
+    CleanlyLoadingComponent component(&engine, QUrl("qrc:///data/jsimport.qml"));
     QScopedPointer<QObject> obj(component.create());
     QVERIFY(!obj.isNull());
     QTRY_COMPARE(obj->property("value").toInt(), 42);
 }
 
+void tst_qmlcachegen::fsScriptImport()
+{
+    QTemporaryDir tempDir;
+    QVERIFY(tempDir.isValid());
+
+    const auto writeTempFile = [&tempDir](const QString &fileName, const char *contents) {
+        QFile f(tempDir.path() + '/' + fileName);
+        const bool ok = f.open(QIODevice::WriteOnly | QIODevice::Truncate);
+        Q_ASSERT(ok);
+        f.write(contents);
+        return f.fileName();
+    };
+
+    const QString testFilePath = writeTempFile(
+            "test.qml",
+            "import QtQml 2.0\n"
+            "import \"test.js\" as ScriptTest\n"
+            "QtObject {\n"
+            "    property int value: ScriptTest.value\n"
+            "}\n");
+
+    const QString scriptFilePath = writeTempFile(
+            "test.js",
+            "var value = 42"
+            );
+
+    QVERIFY(generateCache(scriptFilePath));
+    QVERIFY(generateCache(testFilePath));
+
+    const QString scriptCacheFilePath = scriptFilePath + QLatin1Char('c');
+    QVERIFY(QFile::exists(scriptFilePath));
+
+    {
+        QFile cache(scriptCacheFilePath);
+        QVERIFY(cache.open(QIODevice::ReadOnly));
+        const QV4::CompiledData::Unit *cacheUnit = reinterpret_cast<const QV4::CompiledData::Unit *>(cache.map(/*offset*/0, sizeof(QV4::CompiledData::Unit)));
+        QVERIFY(cacheUnit);
+        QVERIFY(cacheUnit->flags & QV4::CompiledData::Unit::StaticData);
+        QVERIFY(!(cacheUnit->flags & QV4::CompiledData::Unit::PendingTypeCompilation));
+        QCOMPARE(uint(cacheUnit->sourceFileIndex), uint(0));
+    }
+
+    // Remove source code to make sure that when loading succeeds, it is because we loaded
+    // the existing cache files.
+    QVERIFY(QFile::remove(testFilePath));
+    QVERIFY(QFile::remove(scriptFilePath));
+
+    QQmlEngine engine;
+    CleanlyLoadingComponent component(&engine, QUrl::fromLocalFile(testFilePath));
+    QScopedPointer<QObject> obj(component.create());
+    QVERIFY(!obj.isNull());
+    QCOMPARE(obj->property("value").toInt(), 42);
+}
+
+void tst_qmlcachegen::moduleScriptImport()
+{
+    QQmlEngine engine;
+    CleanlyLoadingComponent component(&engine, QUrl("qrc:///data/jsmoduleimport.qml"));
+    QVERIFY2(!component.isError(), qPrintable(component.errorString()));
+    QScopedPointer<QObject> obj(component.create());
+    QVERIFY(!obj.isNull());
+    QTRY_VERIFY(obj->property("ok").toBool());
+
+    QVERIFY(QFile::exists(":/data/script.mjs"));
+    QVERIFY(QFileInfo(":/data/script.mjs").size() > 0);
+
+    {
+        auto componentPrivate = QQmlComponentPrivate::get(&component);
+        QVERIFY(componentPrivate);
+        auto compilationUnit = componentPrivate->compilationUnit->dependentScripts.first()->compilationUnit();
+        QVERIFY(compilationUnit);
+        auto unitData = compilationUnit->unitData();
+        QVERIFY(unitData);
+        QVERIFY(unitData->flags & QV4::CompiledData::Unit::StaticData);
+        QVERIFY(unitData->flags & QV4::CompiledData::Unit::IsESModule);
+
+        QQmlMetaType::CachedUnitLookupError error = QQmlMetaType::CachedUnitLookupError::NoError;
+        const QV4::CompiledData::Unit *unitFromResources = QQmlMetaType::findCachedCompilationUnit(
+                QUrl("qrc:/data/script.mjs"), &error);
+        QVERIFY(unitFromResources);
+
+        QCOMPARE(unitFromResources, compilationUnit->unitData());
+    }
+}
+
+void tst_qmlcachegen::esModulesViaQJSEngine()
+{
+    QJSEngine engine;
+    QJSValue module = engine.importModule(":/data/module.mjs");
+    QJSValue result = module.property("entry").call();
+    QCOMPARE(result.toString(), "ok");
+}
+
 void tst_qmlcachegen::enums()
 {
     QQmlEngine engine;
-    CleanlyLoadingComponent component(&engine, QUrl("qrc:///Enums.qml"));
+    CleanlyLoadingComponent component(&engine, QUrl("qrc:///data/Enums.qml"));
     QScopedPointer<QObject> obj(component.create());
     QVERIFY(!obj.isNull());
     QTRY_COMPARE(obj->property("value").toInt(), 200);
 }
 
+void tst_qmlcachegen::sourceFileIndices()
+{
+    QVERIFY(QFile::exists(":/data/versionchecks.qml"));
+    QVERIFY(QFileInfo(":/data/versionchecks.qml").size() > 0);
+
+    QQmlMetaType::CachedUnitLookupError error = QQmlMetaType::CachedUnitLookupError::NoError;
+    const QV4::CompiledData::Unit *unitFromResources = QQmlMetaType::findCachedCompilationUnit(
+            QUrl("qrc:/data/versionchecks.qml"), &error);
+    QVERIFY(unitFromResources);
+    QVERIFY(unitFromResources->flags & QV4::CompiledData::Unit::PendingTypeCompilation);
+    QCOMPARE(uint(unitFromResources->sourceFileIndex), uint(0));
+}
+
+void tst_qmlcachegen::reproducibleCache_data()
+{
+    QTest::addColumn<QString>("filePath");
+
+    QDir dir(dataDirectory());
+    for (const QString &entry : dir.entryList((QStringList() << "*.qml" << "*.js" << "*.mjs"), QDir::Files)) {
+        QTest::newRow(entry.toUtf8().constData()) << dir.filePath(entry);
+    }
+}
+
+void tst_qmlcachegen::reproducibleCache()
+{
+    QFETCH(QString, filePath);
+
+    QFile file(filePath);
+    QVERIFY(file.exists());
+
+    auto generate = [](const QString &path) {
+        if (!generateCache(path))
+            return QByteArray();
+        QFile generated(path + 'c');
+        [&](){ QVERIFY(generated.open(QIODevice::ReadOnly)); }();
+        const QByteArray result = generated.readAll();
+        generated.remove();
+        return result;
+    };
+
+    const QByteArray contents1 = generate(file.fileName());
+    const QByteArray contents2 = generate(file.fileName());
+    QCOMPARE(contents1, contents2);
+}
+
+void tst_qmlcachegen::parameterAdjustment()
+{
+    QQmlEngine engine;
+    CleanlyLoadingComponent component(&engine, QUrl("qrc:///data/parameterAdjustment.qml"));
+    QScopedPointer<QObject> obj(component.create());
+    QVERIFY(!obj.isNull()); // Doesn't crash
+}
+
+
+void tst_qmlcachegen::inlineComponent()
+{
+    bool ok = generateCache(testFile("inlineComponentWithId.qml"));
+    QVERIFY(ok);
+    QQmlEngine engine;
+    CleanlyLoadingComponent component(&engine, testFileUrl("inlineComponentWithId.qml"));
+    QTest::ignoreMessage(QtMsgType::QtInfoMsg, "42");
+    QScopedPointer<QObject> obj(component.create());
+    QVERIFY(!obj.isNull());
+}
+
+void tst_qmlcachegen::posthocRequired()
+{
+    bool ok = generateCache(testFile("posthocrequired.qml"));
+    QVERIFY(ok);
+    QQmlEngine engine;
+    CleanlyLoadingComponent component(&engine, testFileUrl("posthocrequired.qml"));
+    QScopedPointer<QObject> obj(component.create());
+    QVERIFY(obj.isNull() && component.isError());
+    QVERIFY(component.errorString().contains(QStringLiteral("Required property x was not initialized")));
+}
+
 QTEST_GUILESS_MAIN(tst_qmlcachegen)
 
 #include "tst_qmlcachegen.moc"
