@@ -30,8 +30,9 @@
 #include <private/qqmljsparser_p.h>
 #include <private/qqmljsast_p.h>
 #include <private/qv4codegen_p.h>
-#include <private/qv4value_p.h>
+#include <private/qv4staticvalue_p.h>
 #include <private/qqmlirbuilder_p.h>
+#include <private/qqmljsdiagnosticmessage_p.h>
 
 #include <QtCore/QCoreApplication>
 #include <QtCore/QDir>
@@ -48,6 +49,8 @@
 #include <QtCore/QJsonDocument>
 #include <QtCore/QLibraryInfo>
 
+#include <resourcefilemapper.h>
+
 #include <iostream>
 #include <algorithm>
 
@@ -79,23 +82,24 @@ void printUsage(const QString &appNameIn)
 #endif
     std::wcerr
         << "Usage: " << appName << " -rootPath path/to/app/qml/directory -importPath path/to/qt/qml/directory\n"
-           "       " << appName << " -qmlFiles file1 file2 -importPath path/to/qt/qml/directory\n\n"
+           "       " << appName << " -qmlFiles file1 file2 -importPath path/to/qt/qml/directory\n"
+           "       " << appName << " -qrcFiles file1.qrc file2.qrc -importPath path/to/qt/qml/directory\n\n"
            "Example: " << appName << " -rootPath . -importPath "
         << QDir::toNativeSeparators(qmlPath).toStdWString()
         << '\n';
 }
 
-QVariantList findImportsInAst(QQmlJS::AST::UiHeaderItemList *headerItemList, const QString &code, const QString &path)
+QVariantList findImportsInAst(QQmlJS::AST::UiHeaderItemList *headerItemList, const QString &path)
 {
     QVariantList imports;
 
-    // extract uri and version from the imports (which look like "import Foo.Bar 1.2.3")
+    // Extract uri and version from the imports (which look like "import Foo.Bar 1.2.3")
     for (QQmlJS::AST::UiHeaderItemList *headerItemIt = headerItemList; headerItemIt; headerItemIt = headerItemIt->next) {
         QVariantMap import;
         QQmlJS::AST::UiImport *importNode = QQmlJS::AST::cast<QQmlJS::AST::UiImport *>(headerItemIt->headerItem);
         if (!importNode)
             continue;
-        // handle directory imports
+        // Handle directory imports
         if (!importNode->fileName.isEmpty()) {
             QString name = importNode->fileName.toString();
             import[nameLiteral()] = name;
@@ -116,10 +120,15 @@ QVariantList findImportsInAst(QQmlJS::AST::UiHeaderItemList *headerItemList, con
                 uri = uri->next;
             }
             name.chop(1); // remove trailing "."
+#if QT_VERSION < QT_VERSION_CHECK(6, 0, 0)
+            if (name.startsWith(QLatin1String("QtQuick.Controls")) && name.endsWith(QLatin1String("impl")))
+                continue;
+#endif
             if (!name.isEmpty())
                 import[nameLiteral()] = name;
             import[typeLiteral()] = moduleLiteral();
-            import[versionLiteral()] = code.mid(importNode->versionToken.offset, importNode->versionToken.length);
+            auto versionString = importNode->version ? QString::number(importNode->version->majorVersion) + QLatin1Char('.') + QString::number(importNode->version->minorVersion) : QString();
+            import[versionLiteral()] = versionString;
         }
 
         imports.append(import);
@@ -137,7 +146,7 @@ QVariantMap pluginsForModulePath(const QString &modulePath) {
 
     qmldirFile.open(QIODevice::ReadOnly | QIODevice::Text);
 
-    // a qml import may contain several plugins
+    // A qml import may contain several plugins
     QString plugins;
     QString classnames;
     QStringList dependencies;
@@ -174,7 +183,7 @@ QPair<QString, QString> resolveImportPath(const QString &uri, const QString &ver
 {
     const QLatin1Char dot('.');
     const QLatin1Char slash('/');
-    const QStringList parts = uri.split(dot, QString::SkipEmptyParts);
+    const QStringList parts = uri.split(dot, Qt::SkipEmptyParts);
 
     QString ver = version;
     while (true) {
@@ -206,7 +215,7 @@ QPair<QString, QString> resolveImportPath(const QString &uri, const QString &ver
             }
         }
 
-        // remove the last version digit; stop if there are none left
+        // Remove the last version digit; stop if there are none left
         if (ver.isEmpty())
             break;
 
@@ -276,7 +285,7 @@ QVariantList findQmlImportsInQmlCode(const QString &filePath, const QString &cod
         }
         return QVariantList();
     }
-    return findImportsInAst(parser.ast()->headers, code, filePath);
+    return findImportsInAst(parser.ast()->headers, filePath);
 }
 
 // Scan a single qml file for import statements
@@ -426,7 +435,7 @@ QVariantList findQmlImportsInDirectory(const QString &qmlDir)
         if (std::find_if(blacklist.cbegin(), blacklist.cend(), pathStartsWith(path)) != blacklist.cend())
             continue;
 
-        // skip obvious build output directories
+        // Skip obvious build output directories
         if (path.contains(QLatin1String("Debug-iphoneos")) || path.contains(QLatin1String("Release-iphoneos")) ||
             path.contains(QLatin1String("Debug-iphonesimulator")) || path.contains(QLatin1String("Release-iphonesimulator"))
 #ifdef Q_OS_WIN
@@ -455,30 +464,29 @@ QSet<QString> importModulePaths(const QVariantList &imports) {
     return ret;
 }
 
-// Find Qml Imports Recursively from a root set of qml files.
+// Find qml imports recursively from a root set of qml files.
 // The directories in qmlDirs are searched recursively.
 // The files in qmlFiles parsed directly.
 QVariantList findQmlImportsRecursively(const QStringList &qmlDirs, const QStringList &scanFiles)
 {
     QVariantList ret;
 
-    // scan all app root qml directories for imports
+    // Scan all app root qml directories for imports
     for (const QString &qmlDir : qmlDirs) {
         QVariantList imports = findQmlImportsInDirectory(qmlDir);
         ret = mergeImports(ret, imports);
     }
 
-    // scan app qml files for imports
+    // Scan app qml files for imports
     for (const QString &file : scanFiles) {
         QVariantList imports = findQmlImportsInFile(file);
         ret = mergeImports(ret, imports);
     }
 
-
-    // get the paths to theimports found in the app qml
+    // Get the paths to the imports found in the app qml
     QSet<QString> toVisit = importModulePaths(ret);
 
-    // recursivly scan for import dependencies.
+    // Recursively scan for import dependencies.
     QSet<QString> visited;
     while (!toVisit.isEmpty()) {
         QString qmlDir = *toVisit.begin();
@@ -495,6 +503,36 @@ QVariantList findQmlImportsRecursively(const QStringList &qmlDirs, const QString
     return ret;
 }
 
+
+QString generateCmakeIncludeFileContent(const QVariantList &importList) {
+    // The function assumes that "list" is a QVariantList with 0 or more QVariantMaps, where
+    // each map contains QString -> QVariant<QString> mappings. This matches with the structure
+    // that qmake parses for static qml plugin auto imporitng.
+    // So: [ {"a": "a","b": "b"}, {"c": "c"} ]
+    QString content;
+    QTextStream s(&content);
+    int importsCount = 0;
+    for (const QVariant &importVariant: importList) {
+        if (static_cast<QMetaType::Type>(importVariant.userType()) == QMetaType::QVariantMap) {
+            s << QStringLiteral("set(qml_import_scanner_import_") << importsCount
+              << QStringLiteral(" \"");
+
+            const QMap<QString, QVariant> &importDict = importVariant.toMap();
+            for (auto it = importDict.cbegin(); it != importDict.cend(); ++it) {
+                s << it.key().toUpper() << QLatin1Char(';')
+                  << it.value().toString() << QLatin1Char(';');
+            }
+            s << QStringLiteral("\")\n");
+            ++importsCount;
+        }
+    }
+    if (importsCount >= 0) {
+        content.prepend(QString(QStringLiteral("set(qml_import_scanner_imports_count %1)\n"))
+               .arg(importsCount));
+    }
+    return content;
+}
+
 } // namespace
 
 int main(int argc, char *argv[])
@@ -511,6 +549,8 @@ int main(int argc, char *argv[])
     QStringList qmlRootPaths;
     QStringList scanFiles;
     QStringList qmlImportPaths;
+    QStringList qrcFiles;
+    bool generateCmakeContent = false;
 
     int i = 1;
     while (i < args.count()) {
@@ -535,6 +575,10 @@ int main(int argc, char *argv[])
             if (i >= args.count())
                 std::cerr << "-importPath requires an argument\n";
             argReceiver = &qmlImportPaths;
+        } else if (arg == QLatin1String("-cmake-output")) {
+             generateCmakeContent = true;
+        } else if (arg == QLatin1String("-qrcFiles")) {
+            argReceiver = &qrcFiles;
         } else {
             std::cerr << qPrintable(appName) << ": Invalid argument: \""
                 << qPrintable(arg) << "\"\n";
@@ -556,13 +600,23 @@ int main(int argc, char *argv[])
         }
     }
 
+    if (!qrcFiles.isEmpty())
+        scanFiles << ResourceFileMapper(qrcFiles).qmlCompilerFiles(ResourceFileMapper::FileOutput::AbsoluteFilePath);
+
     g_qmlImportPaths = qmlImportPaths;
 
     // Find the imports!
     QVariantList imports = findQmlImportsRecursively(qmlRootPaths, scanFiles);
 
-    // Convert to JSON
-    QByteArray json = QJsonDocument(QJsonArray::fromVariantList(imports)).toJson();
-    std::cout << json.constData() << std::endl;
+    QByteArray content;
+    if (generateCmakeContent) {
+        // Convert to CMake code
+        content = generateCmakeIncludeFileContent(imports).toUtf8();
+    } else {
+        // Convert to JSON
+        content = QJsonDocument(QJsonArray::fromVariantList(imports)).toJson();
+    }
+
+    std::cout << content.constData() << std::endl;
     return 0;
 }
